# LUT Gamma Converter Ver. 1.0
# (c)2024 Ohnishi, Yasuo
# 2024.6
from ij import IJ
from ij import WindowManager as WM
from ij.io import OpenDialog
from ij.process import ImageConverter
from ij.plugin import ChannelSplitter, RGBStackMerge
from ij.plugin import Duplicator
from ij.process import LUT
from ij.gui import GenericDialog
from loci.plugins import BF
from java.awt.event import AdjustmentListener, ItemListener
from jarray import zeros, array
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#--- image type detection
def imageTypeDetection(imp):
    StackSize = imp.getStackSize()
    BitDepth = imp.getBitDepth()
    return StackSize, BitDepth

#----load image-------
def load_image():
    bioImport = False
    imgCurrent = WM.getCurrentImage()
    if not imgCurrent:
        bioImport = True
        IJ.showMessage("New file will be opened with bio-format importer,\n and it will reset LUT of the file.")
        od = OpenDialog("Choose a file")
        folder = od.getDirectory()
        filename = od.getFileName()
    # Always Open file with Bio-format Plugin
        if filename:
            imps = BF.openImagePlus(folder + filename)
            imp0 = imps[0] #imp0: Original Image
            imp0.show()
        else:
            IJ.showMessage("File is not selected.")
    else: # target image is already open
        imp0 = imgCurrent
        filename = imp0.getTitle()
        try:
            folder = imp0.getOriginalFileInfo().directory
        except:
            folder = IJ.getDirectory("Output_directory")
    return imp0, filename, folder, bioImport

#-----------------------------
class imagePreviewer(AdjustmentListener, ItemListener):
    def __init__(self, imp, impPreview, nfield_Gamma, choice_process):
        self.n = 0
        self.imp = imp
        self.imp.setTitle("self.imp")
        self.choice_process = choice_process
        self.StackSize = imp.getStackSize()
        self.impPreview = impPreview
        self.impPreview.setTitle("self.impPreview")
        self.impPreview.show()
        WM.getActiveWindow().toFront()
        self.nfield_Gamma = nfield_Gamma

    def adjustmentValueChanged(self, event):
        if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
            return
        self.execution()

    def itemStateChanged(self, event):
        if event.getStateChange() == event.SELECTED:
            self.execution()

    def execution(self):
        #ip = self.imp.getProcessor()
        Gamma = float(self.nfield_Gamma.getText())
        process = self.choice_process.getSelectedItem()
        if process == "adjust existing Lut":
            process = "adjust"
        impPrv2 = Duplicator().run(self.imp)
        self.imp.hide()
        if self.StackSize == 1:
            print ("Call TRCencode Ch1")
            lut = TRCencode(self.imp, Gamma, 3, process)
            self.impPreview.getProcessor().setColorModel(lut)
            self.impPreview.updateAndDraw()
            self.impPreview.show()
        else:
            impsCh1 = ChannelSplitter.split(impPrv2)
            for i in range(3):
                if process == "encode" or i == 1:
                    lut = TRCencode(impsCh1[i], Gamma, i, process)
                    impsCh1[i].getProcessor().setColorModel(lut)
            imp2 = RGBStackMerge.mergeChannels(impsCh1, True)
            self.impPreview.hide()
            self.impPreview.close()

            self.impPreview = imp2
            self.n = self.n + 1
            self.impPreview.setTitle("imp2")
            self.impPreview.show()
            for i in range(3):
                impsCh1[i].hide()
                impsCh1[i].close()
        impPrv2.hide()
        impPrv2.close()

#-----------------------------
def get_Choice(impPrv, bioImport):
        # dialog for Gamma value specifying
    gd = GenericDialog("Apply Gamma encoded LUT")
    if bioImport == True:
        gd.addMessage("To adjust existing Lut \n open the file in advance")
    if bioImport == False:
        processChoice = ["encode", "adjust existing Lut"]
    else:
        processChoice = ["encode"]
    gd.addChoice("Select applying process", processChoice, processChoice[0])
    gd.addSlider("Gamma", 0.0, 5.0, 1.0, 0.01)  #slider 0 / NField 1
    choice_process = gd.getChoices().get(0)
    slider_Gamma  = gd.getSliders().get(0) # slider 0
    nfield_Gamma  = gd.getNumericFields().get(0) # NField #0

    impCanvas = impPrv.duplicate()
    impPrv.setTitle("impPrv")
    impPrv.setTitle("impCanvas")
    previewer =  imagePreviewer(impPrv, impCanvas, nfield_Gamma, choice_process)
    choice_process.addItemListener(previewer)
    slider_Gamma.addAdjustmentListener(previewer) # set listener for upper slider


    #---- show the dialog ----
    gd.showDialog()

    process =  gd.getNextChoice()
    if process == "adjust existing Lut":
        process = "adjust"
    Gamma = gd.getNextNumber()
    if bioImport == True and process == "adjust":
        IJ.showMessage\
            ("To adjust existing Lut, you should open image in advance.")
        process = "cancel"
    if gd.wasCanceled():
        process = "cancel"
    impPrv.hide()
    impPrv.close()
    impCanvas.hide()
    impCanvas.close()
    choice_process.removeItemListener(previewer)
    slider_Gamma.removeAdjustmentListener(previewer) # set listener for upper slider


    return process, Gamma

#--------------------------
def TRCencode(imp, Gamma, Ch, process):
    # if imp is mono channel: Ch == 3
    lut0 = []
    channel = zeros(256, 'b')
    channel0 = zeros(256, 'b')
    channelR = zeros(256, 'b')
    channelG = zeros(256, 'b')
    channelB = zeros(256, 'b')

    #channela =  array([[0.0] * 256 * 3], 'b')
    print ("Stack Channel Number: " + str(Ch))
    if Ch == 3 or process == "adjust":
        NofCh =1
    else:
        NofCh =3
    for i in range(3):
        #print ("Lut Channel Number: " + str(i))
        #print ("Int. index", "Byte index", \
        #       "Data Type", "Org. Value", "Value changed", "Byte value")
        if (process == "encode" and i == Ch) or NofCh == 1:
            for j in range(256):
                if process == "encode": ## TRC encoding
                    v = j
                elif i == 0: ## adjustment of lut
                    v = imp.getProcessor().getLut().getRed(j)
                elif i == 1:
                    v = imp.getProcessor().getLut().getGreen(j)
                elif i == 2:
                    v = imp.getProcessor().getLut().getBlue(j)
                if Gamma == 0.0:
                    Gamma = 0.001
                dataTmp =  int((((v / 255.0) ** (1.0 / Gamma)) * 255))
                if j < 128:
                    j2 = j
                else:
                    j2 = j - 256
                if dataTmp < 128:
                    dataTmp1 = dataTmp
                else:
                    dataTmp1 = dataTmp - 256
                #if j < 10:
                #    print (j, j2, type(dataTmp), v, dataTmp, dataTmp1)
                channel[j2] = dataTmp1
                if NofCh == 1: ## Lut adjustment for mono channel or RGB
                    if i == 0:
                        channelR[j2] = dataTmp1
                    elif i == 1:
                        channelG[j2] = dataTmp1
                    elif i == 2:
                        channelB[j2] = dataTmp1

    if process == "adjust": ## mono channel or RGB Lut adjustment
        lut = LUT(8, 256, channelR, channelG, channelB)
    elif Ch == 3: ## mono channel TRC encoding
        lut = LUT(8, 256, channel, channel, channel)
    elif Ch == 0: # RGB image TRC encoding
        lut = LUT(8, 256, channel, channel0, channel0)
    elif Ch == 1:
        lut = LUT(8, 256, channel0, channel, channel0)
    elif Ch == 2:
        lut = LUT(8, 256, channel0, channel0, channel)
    return lut

#------------------------
def set_Lmax(imp):
    # set Maximum value of Lightness
    bitdepth = imp.getBitDepth()
    if bitdepth == 32:
        L_max = imp.getDisplayRangeMax()
        if int(L_max) > 65536:
            L_max = 16777215
        elif int(L_max) > 256:
            L_max = 65535
        elif int(L_max) > 1.01:
            L_max = 255
        else:
            L_max = 1.0
    elif bitdepth == 16:
        L_max = 65535
    else: # 8 or 24 bit
        L_max = 255
    return L_max
#--------------------------------
def make_preview(imp):
    """ making preview image """
    impprv = Duplicator().run(imp) # impprv: preview imp
    impprv.setDisplayRange(0.0, set_Lmax(imp))
    scale = 1280.0 / impprv.getWidth()
                    # preview image scaling value (width = 1200px fixed)
    new_height = impprv.getHeight() * scale # preview image height
    new_height = int(new_height)
    impprv1 = impprv.resize(1280, new_height, "none")
    return impprv1

#---  code for main ---
imp0, filename, folder, bioImport = load_image()
StackSize, BitDepth = imageTypeDetection(imp0)
if StackSize > 3:
    IJ.showMessage("StackSize is " + str(StackSize))
    StackSize = 3
if BitDepth == 24: # color processor
    BitDepth = 8
impPrv = make_preview(imp0)
imp0.hide()
impPrv.show()
process, Gamma = get_Choice(impPrv, bioImport)
impPrv.hide()
impPrv.close()
if process != "cancel":
    if filename:
        IJ.log("folder: " + folder)
        IJ.log("filename: " + filename)
        IJ.log("StackSize: " + str(StackSize))
        IJ.log("BitDepth: " + str(BitDepth))
    imp1 = Duplicator().run(imp0)
    if process != "cancel":
        if StackSize == 1:
            lut0 = imp0.getProcessor().getLut()
            lut = TRCencode(imp0, Gamma, 3, process)
            imp1.getProcessor().setColorModel(lut)
            imp2 = imp1
        else:
            impsCh1 = ChannelSplitter.split(imp1)
            for i in range(3):
                #if process == "adjust" and i > 0:
                #    impsCh1[i].getProcessor().setLut(None)
                if process == "encode" or i == 1:
                    lut0 = impsCh1[i].getProcessor().getLut()
                    lut = TRCencode(impsCh1[i], Gamma, i, process)
                    impsCh1[i].getProcessor().setColorModel(lut)
            imp2 = RGBStackMerge.mergeChannels(impsCh1, True)
        IJ.saveAsTiff(imp2, folder + filename +"_Gamma_" + \
                       str(Gamma) + "_" + process +"d.tif")
    imp0.close()
    imp2.show()
    if StackSize > 1:
        impClose = WM.getImage("imp2")
        impClose.hide()
        impClose.close()
    IJ.showMessage("TRC Conversion Finished")
    if StackSize > 1:
        for i in range(StackSize):
            impsCh1[i].close()
    if StackSize > 1:
        imp1.close()
else:
    IJ.showMessage("Job canceled")
    imp0.close()

#---

